if not ShockAndAwe_Bars then return end

local L = LibStub("AceLocale-3.0"):GetLocale("ShockAndAwe_Bars")
local media = LibStub:GetLibrary("LibSharedMedia-3.0");
local LBF = LibStub("LibButtonFacade", true)
local LBFgroup = nil
local _, _, _, clientVersion = GetBuildInfo()

--if LBF then
--	local LBFgroup = LBF:Group("ShockAndAwe_Bars")
--end
ShockAndAwe_Bars.PriorityFrame = CreateFrame("Button", "SAA_PriorityFrame", UIParent)
ShockAndAwe_Bars.PriorityFrame.cooldown = CreateFrame("Cooldown", "SAA_PriorityFrameCooldown", ShockAndAwe_Bars.PriorityFrame)

---------------------------
-- Local Variables
---------------------------

ShockAndAwe_Bars.priorityTable = {}
ShockAndAwe_Bars.priorityTable.name = {}
ShockAndAwe_Bars.priorityTable.icon = {}
ShockAndAwe_Bars.priorityTable.test = {}

--------------------------
--  Helper Functions
--------------------------
-- returns : true if spellname is available in less than priority.cooldown
--                 or if spellname is currently activated
--           , remaining cooldown, 0 if ready to be cast
function ShockAndAwe_Bars:SpellAvailable(spellname)
	if ShockAndAwe_Bars.db.char.priority.hideImmune and self:IsTargetImmune(spellname) then
		return false
	end
	local start, duration = GetSpellCooldown(spellname)
	if start then
		local timeleft = start + duration - GetTime()
		return timeleft <= ShockAndAwe_Bars.db.char.priority.cooldown, duration, timeleft
	end
	return false, 999
end

function ShockAndAwe_Bars:GCDAvailable()
	local startTime, gcdDuration = GetSpellCooldown(ShockAndAwe_Bars.constants["Lightning Bolt"])
	local gcdTime = startTime + gcdDuration - GetTime()
	return gcdTime <= ShockAndAwe_Bars.db.char.priority.cooldown 
end

function ShockAndAwe_Bars:GetDebuffInfo(debuff)
	local index = 1
	while UnitDebuff("target", index) do
		local name, _, _, count, _, _, debuffExpires, unitCaster = UnitDebuff("target", index)
		local isMine = unitCaster == "player" 
		if name == debuff and isMine then 
			local duration = debuffExpires - GetTime()
			return duration > ShockAndAwe_Bars.db.char.priority.cooldown, duration, count, debuffExpires
		end
		index = index + 1
	end
	return false, 0, 0, 0
end

function ShockAndAwe_Bars:GetBuffInfo(buff)
	local index = 1
	while UnitBuff("player", index) do
		local name, _, _, count, _, _, buffExpires = UnitBuff("player", index)
		if name == buff then 
			local duration = buffExpires - GetTime()
			return duration > ShockAndAwe_Bars.db.char.priority.cooldown, duration, count
		end
		index = index + 1
	end
	return false, 0, 0
end

function ShockAndAwe_Bars:IsTargetImmune(spell)
	target = UnitName("target") or ""
	if spell and ShockAndAwe_Bars.db.char.immuneTargets[target.."_"..spell] then
		return true
	end
	return false
end

function ShockAndAwe_Bars:FireTotemNeeded(spell)
	if ShockAndAwe_Bars:GCDAvailable() then
		-- only bother checking for totem availabilty if we have a GCD
		local _, totemname, start, duration = GetTotemInfo(1)
		if totemname == nil or totemname == "" then -- no fire totem deployed so recommend it be dropped
			return true
		end
		if start and strmatch(totemname, spell) then
			local timeleft = start + duration - GetTime()
			return timeleft <= ShockAndAwe_Bars.db.char.priority.cooldown, duration
		end
	end
	return false, 999
end

------------------
-- Priorities
------------------

function ShockAndAwe_Bars:SetPriorityTable()
	ShockAndAwe_Bars.priorityTable.name["sr"] = ShockAndAwe_Bars.constants["Shamanistic Rage"]
	ShockAndAwe_Bars.priorityTable.icon["sr"] = ShockAndAwe_Bars.constants["Shamanistic Rage Icon"]
	ShockAndAwe_Bars.priorityTable.test["sr"] = 
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Shamanistic Rage"])
		end 

	ShockAndAwe_Bars.priorityTable.name["sr_boss"] = L["Shamanistic Rage on Boss"]
	ShockAndAwe_Bars.priorityTable.icon["sr_boss"] = ShockAndAwe_Bars.constants["Shamanistic Rage Icon"]
	ShockAndAwe_Bars.priorityTable.test["sr_boss"] = 
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Shamanistic Rage"]) and UnitClassification("target") == "worldboss"
		end 

	ShockAndAwe_Bars.priorityTable.name["sw"] = ShockAndAwe_Bars.constants["Feral Spirit"]
	ShockAndAwe_Bars.priorityTable.icon["sw"] = ShockAndAwe_Bars.constants["Feral Spirit Icon"]
	ShockAndAwe_Bars.priorityTable.test["sw"] = 
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Feral Spirit"])
		end 

	ShockAndAwe_Bars.priorityTable.name["sw_boss"] = L["Feral Spirit on Boss"]
	ShockAndAwe_Bars.priorityTable.icon["sw_boss"] = ShockAndAwe_Bars.constants["Feral Spirit Icon"]
	ShockAndAwe_Bars.priorityTable.test["sw_boss"] = 
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Feral Spirit"]) and UnitClassification("target") == "worldboss"
		end 

	ShockAndAwe_Bars.priorityTable.name["mw5_lb"] = L["MW5 Lightning Bolt"]
	ShockAndAwe_Bars.priorityTable.icon["mw5_lb"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw5_lb"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks == 5 
		end 
		
	ShockAndAwe_Bars.priorityTable.name["lb_as"] = L["Lightning Bolt if Ancestral Swiftness is up"]
	ShockAndAwe_Bars.priorityTable.icon["lb_as"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["lb_as"] = 
		function () 
		    local start, duration, enabled = GetSpellCooldown("Ancestral Swiftness")
			if enabled == 0 then
			  return true
			else
			  return false
			end
		end
		
	ShockAndAwe_Bars.priorityTable.name["cl_as"] = L["Chain Lightning if Ancestral Swiftness is up"]
	ShockAndAwe_Bars.priorityTable.icon["cl_as"] = ShockAndAwe_Bars.constants["Chain Lightning Icon"]
	ShockAndAwe_Bars.priorityTable.test["cl_as"] = 
		function () 
		    local start, duration, enabled = GetSpellCooldown("Ancestral Swiftness")
			if enabled == 0 then
			  return true
			else
			  return false
			end
		end

	ShockAndAwe_Bars.priorityTable.name["mw4_lb"] = L["MW4 Lightning Bolt"]
	ShockAndAwe_Bars.priorityTable.icon["mw4_lb"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw4_lb"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks >= 4 
		end 

	ShockAndAwe_Bars.priorityTable.name["mw3_lb"] = L["MW3 Lightning Bolt"]
	ShockAndAwe_Bars.priorityTable.icon["mw3_lb"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw3_lb"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks >= 3 
		end 

	ShockAndAwe_Bars.priorityTable.name["mw5_cl"] = L["MW5 Chain Lightning"]
	ShockAndAwe_Bars.priorityTable.icon["mw5_cl"] = ShockAndAwe_Bars.constants["Chain Lightning Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw5_cl"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks == 5 and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Chain Lightning"])
		end 

	ShockAndAwe_Bars.priorityTable.name["mw4_cl"] = L["MW4 Chain Lightning"]
	ShockAndAwe_Bars.priorityTable.icon["mw4_cl"] = ShockAndAwe_Bars.constants["Chain Lightning Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw4_cl"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks >= 4 and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Chain Lightning"])
		end 

	ShockAndAwe_Bars.priorityTable.name["mw3_cl"] = L["MW3 Chain Lightning"]
	ShockAndAwe_Bars.priorityTable.icon["mw3_cl"] = ShockAndAwe_Bars.constants["Chain Lightning Icon"]
	ShockAndAwe_Bars.priorityTable.test["mw3_cl"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks >= 3 and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Chain Lightning"])
		end 

	ShockAndAwe_Bars.priorityTable.name["ss"] = ShockAndAwe_Bars.constants["Stormstrike"]
	ShockAndAwe_Bars.priorityTable.icon["ss"] = ShockAndAwe_Bars.constants["Stormstrike Icon"]
	ShockAndAwe_Bars.priorityTable.test["ss"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Stormstrike"])
		end 

	ShockAndAwe_Bars.priorityTable.name["ss_0"] = L["Stormstrike if no debuff"]
	ShockAndAwe_Bars.priorityTable.icon["ss_0"] = ShockAndAwe_Bars.constants["Stormstrike Icon"]
	ShockAndAwe_Bars.priorityTable.test["ss_0"] =  
	function () 
		return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Stormstrike"]) and not ShockAndAwe_Bars:GetDebuffInfo(ShockAndAwe_Bars.constants["Stormstrike"])
	end 

	ShockAndAwe_Bars.priorityTable.name["es"] = ShockAndAwe_Bars.constants["Earth Shock"]
	ShockAndAwe_Bars.priorityTable.icon["es"] = ShockAndAwe_Bars.constants["Earth Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["es"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Earth Shock"])
		end 

	ShockAndAwe_Bars.priorityTable.name["es_ss"] = L["Earth Shock if SS"]
	ShockAndAwe_Bars.priorityTable.icon["es_ss"] = ShockAndAwe_Bars.constants["Earth Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["es_ss"] =  
		function () 
			return ShockAndAwe_Bars:GetDebuffInfo(ShockAndAwe_Bars.constants["Stormstrike"]) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Earth Shock"])
		end 

	ShockAndAwe_Bars.priorityTable.name["fs"] = ShockAndAwe_Bars.constants["Flame Shock"]
	ShockAndAwe_Bars.priorityTable.icon["fs"] = ShockAndAwe_Bars.constants["Flame Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["fs"] =  
		function () 
			local debuffPresent, duration = ShockAndAwe_Bars:GetDebuffInfo(ShockAndAwe_Bars.constants["Flame Shock"])
			local ticksLeft = 6 * duration / (ShockAndAwe_Bars.db.char.priority.FSDotMax or 18)
			return (not debuffPresent or ticksLeft <= ShockAndAwe_Bars.db.char.priority.fsticksleft) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Flame Shock"]) 
		end

	ShockAndAwe_Bars.priorityTable.name["fs_boss"] = L["Flame Shock on Boss"]
	ShockAndAwe_Bars.priorityTable.icon["fs_boss"] = ShockAndAwe_Bars.constants["Flame Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["fs_boss"] =  
		function () 
			local debuffPresent, duration = ShockAndAwe_Bars:GetDebuffInfo(ShockAndAwe_Bars.constants["Flame Shock"])
			local ticksLeft = 6 * duration / (ShockAndAwe_Bars.db.char.priority.FSDotMax or 18)
			return (not debuffPresent or ticksLeft <= ShockAndAwe_Bars.db.char.priority.fsticksleft) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Flame Shock"]) and UnitClassification("target") == "worldboss"
		end 

	ShockAndAwe_Bars.priorityTable.name["fn"] = ShockAndAwe_Bars.constants["Fire Nova"]
	ShockAndAwe_Bars.priorityTable.icon["fn"] = ShockAndAwe_Bars.constants["Fire Nova Icon"]
	ShockAndAwe_Bars.priorityTable.test["fn"] =  
		function () 
			local FSDotTime = ShockAndAwe_Bars:getFSDotTime()			
			if not FSDotTime or FSDotTime < GetTime() then -- no Flame Shock active so cannot use Fire Nova
				return false
			end
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Fire Nova"]) 
		end 

	ShockAndAwe_Bars.priorityTable.name["fe"] = ShockAndAwe_Bars.constants["Fire Elemental Totem"]
	ShockAndAwe_Bars.priorityTable.icon["fe"] = ShockAndAwe_Bars.constants["Fire Elemental Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["fe"] =  
		function () 
			return ShockAndAwe_Bars:FireTotemNeeded(ShockAndAwe_Bars.constants["Fire Elemental Totem"]) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Fire Elemental Totem"]) 
		end 

	ShockAndAwe_Bars.priorityTable.name["fe_boss"] = L["Fire Elemental on Boss"]
	ShockAndAwe_Bars.priorityTable.icon["fe_boss"] = ShockAndAwe_Bars.constants["Fire Elemental Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["fe_boss"] =  
		function () 
			return ShockAndAwe_Bars:FireTotemNeeded(ShockAndAwe_Bars.constants["Fire Elemental Totem"]) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Fire Elemental Totem"]) and UnitClassification("target") == "worldboss"
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ll"] = ShockAndAwe_Bars.constants["Lava Lash"]
	ShockAndAwe_Bars.priorityTable.icon["ll"] = ShockAndAwe_Bars.constants["Lava Lash Icon"]
	ShockAndAwe_Bars.priorityTable.test["ll"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Lava Lash"])
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ll_sf5"] = L["Lava Lash 5 Searing Flames"]
	ShockAndAwe_Bars.priorityTable.icon["ll_sf5"] = ShockAndAwe_Bars.constants["Lava Lash Icon"]
	ShockAndAwe_Bars.priorityTable.test["ll_sf5"] =  
		function () 
			local _, _, sfstacks = ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Searing Flames"])
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Lava Lash"]) and sfstacks == 5
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ll_sf4"] = L["Lava Lash 4 Searing Flames"]
	ShockAndAwe_Bars.priorityTable.icon["ll_sf4"] = ShockAndAwe_Bars.constants["Lava Lash Icon"]
	ShockAndAwe_Bars.priorityTable.test["ll_sf4"] =  
		function () 
			local _, _, sfstacks = ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Searing Flames"])
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Lava Lash"]) and sfstacks >= 4
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ll_sf3"] = L["Lava Lash 3 Searing Flames"]
	ShockAndAwe_Bars.priorityTable.icon["ll_sf3"] = ShockAndAwe_Bars.constants["Lava Lash Icon"]
	ShockAndAwe_Bars.priorityTable.test["ll_sf3"] =  
		function () 
			local _, _, sfstacks = ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Searing Flames"])
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Lava Lash"]) and sfstacks >= 3
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ue"] = ShockAndAwe_Bars.constants["Unleash Elements"]
	ShockAndAwe_Bars.priorityTable.icon["ue"] = ShockAndAwe_Bars.constants["Unleash Elements Icon"]
	ShockAndAwe_Bars.priorityTable.test["ue"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Unleash Elements"])
		end 
		
	ShockAndAwe_Bars.priorityTable.name["ue_uf"] = L["Unleash Elements if Unleashed Fury is known"]
	ShockAndAwe_Bars.priorityTable.icon["ue_uf"] = ShockAndAwe_Bars.constants["Unleash Elements Icon"]
	ShockAndAwe_Bars.priorityTable.test["ue_uf"] =  
		function () 	
            -- Unleash Elements if spec into Unleashed Fury		
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Unleash Elements"]) and IsSpellKnown(117012)
		end
		
	ShockAndAwe_Bars.priorityTable.name["eb"] = ShockAndAwe_Bars.constants["Elemental Blast"]
	ShockAndAwe_Bars.priorityTable.icon["eb"] = ShockAndAwe_Bars.constants["Elemental Blast Icon"]
	ShockAndAwe_Bars.priorityTable.test["eb"] =  
		function () 			
			return IsSpellKnown(117014) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Elemental Blast"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["as_mw<2"] = L["Ancestral Swiftness if MW < 2"]
	ShockAndAwe_Bars.priorityTable.icon["as_mw<2"] = ShockAndAwe_Bars.constants["Ancestral Swiftness Icon"]
	ShockAndAwe_Bars.priorityTable.test["as_mw<2"] =  
		function () 			
			return IsSpellKnown(16188) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Ancestral Swiftness"]) and ShockAndAwe_Bars.db.char.msstacks < 2
		end
		
	ShockAndAwe_Bars.priorityTable.name["as"] = ShockAndAwe_Bars.constants["Ancestral Swiftness"]
	ShockAndAwe_Bars.priorityTable.icon["as"] = ShockAndAwe_Bars.constants["Ancestral Swiftness Icon"]
	ShockAndAwe_Bars.priorityTable.test["as"] =  
		function () 			
			return IsSpellKnown(16188) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Ancestral Swiftness"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["fs_uef"] = L["Flame Shock if Unleash Flame"]
	ShockAndAwe_Bars.priorityTable.icon["fs_uef"] = ShockAndAwe_Bars.constants["Flame Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["fs_uef"] = 
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Flame Shock"]) and ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Unleash Flame"])
		end 
		
	ShockAndAwe_Bars.priorityTable.name["st"] = ShockAndAwe_Bars.constants["Searing Totem"]
	ShockAndAwe_Bars.priorityTable.icon["st"] = ShockAndAwe_Bars.constants["Searing Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["st"] =  
		function () 
			local timeleft = ShockAndAwe_Bars.db.char.priority.firetotemtime - GetTime()
			return (ShockAndAwe_Bars:FireTotemNeeded(ShockAndAwe_Bars.constants["Searing Totem"]) or timeleft < 1.64 * ShockAndAwe_Bars.db.char.priority.searingticks) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Searing Totem"])
		end 

	ShockAndAwe_Bars.priorityTable.name["mt"] = ShockAndAwe_Bars.constants["Magma Totem"]
	ShockAndAwe_Bars.priorityTable.icon["mt"] = ShockAndAwe_Bars.constants["Magma Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["mt"] =  
		function () 
			local timeleft = ShockAndAwe_Bars.db.char.priority.firetotemtime - GetTime()
			return (ShockAndAwe_Bars:FireTotemNeeded(ShockAndAwe_Bars.constants["Magma Totem"]) or timeleft < 2 * ShockAndAwe_Bars.db.char.priority.magmaticks) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Magma Totem"])
		end 

	ShockAndAwe_Bars.priorityTable.name["mt_0"] = L["Magma Totem if expired"]
	ShockAndAwe_Bars.priorityTable.icon["mt_0"] = ShockAndAwe_Bars.constants["Magma Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["mt_0"] =  
		function () 
			return ShockAndAwe_Bars:FireTotemNeeded(ShockAndAwe_Bars.constants["Magma Totem"]) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Magma Totem"])
		end

	ShockAndAwe_Bars.priorityTable.name["ls_0"] = L["Lightning Shield if expired"]
	ShockAndAwe_Bars.priorityTable.icon["ls_0"] = ShockAndAwe_Bars.constants["Lightning Shield Icon"]
	ShockAndAwe_Bars.priorityTable.test["ls_0"] =  
		function () 
			local orbs = ShockAndAwe_Bars:GetShieldInfo()
			return orbs == 0 and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Lightning Shield"])
		end

	ShockAndAwe_Bars.priorityTable.name["mw5_hs"] = L["MW5 Healing Surge"]
	ShockAndAwe_Bars.priorityTable.test["mw5_hs"] = 
		function () 
			if ShockAndAwe_Bars.db.char.priority.chingroup == true and GetNumGroupMembers() > 0 then
				ShockAndAwe_Bars.priorityTable.icon["mw5_hs"] = ShockAndAwe_Bars.constants["Chain Heal Icon"]
			else
				ShockAndAwe_Bars.priorityTable.icon["mw5_hs"] = ShockAndAwe_Bars.constants["Healing Surge Icon"]
			end
			local healthpercent = 100 * UnitHealth("player") / UnitHealthMax("player")
			return ShockAndAwe_Bars.db.char.msstacks == 5 and healthpercent <= ShockAndAwe_Bars.db.char.priority.hwhealth
		end
		
	ShockAndAwe_Bars.priorityTable.name["mw4_hs"] = L["MW4 Healing Surge"]
	ShockAndAwe_Bars.priorityTable.test["mw4_hs"] = 
		function () 
			if ShockAndAwe_Bars.db.char.priority.chingroup == true and GetNumGroupMembers() > 0 then
				ShockAndAwe_Bars.priorityTable.icon["mw4_hs"] = ShockAndAwe_Bars.constants["Chain Heal Icon"]
			else
				ShockAndAwe_Bars.priorityTable.icon["mw4_hs"] = ShockAndAwe_Bars.constants["Healing Surge Icon"]
			end
			local healthpercent = 100 * UnitHealth("player") / UnitHealthMax("player")
			return ShockAndAwe_Bars.db.char.msstacks >= 4 and healthpercent <= ShockAndAwe_Bars.db.char.priority.hwhealth
		end
		
	ShockAndAwe_Bars.priorityTable.name["shield"] = ShockAndAwe_Bars.constants["Water Shield"] .. "/" .. ShockAndAwe_Bars.constants["Lightning Shield"]
	ShockAndAwe_Bars.priorityTable.test["shield"] = 
		function ()
			local orbs, shield, shieldExpires = ShockAndAwe_Bars:GetShieldInfo()
			local timeleft = shieldExpires - GetTime()
			if timeleft <= ShockAndAwe_Bars.db.char.priority.cooldown or orbs <= ShockAndAwe_Bars.db.char.priority.shieldorbs then
				local manapercent = 100 * UnitMana("player") / UnitManaMax("player")
				if manapercent <= ShockAndAwe_Bars.db.char.priority.wsmana then
					ShockAndAwe_Bars.priorityTable.icon["shield"] = ShockAndAwe_Bars.constants["Water Shield Icon"]
					return true
				else
					ShockAndAwe_Bars.priorityTable.icon["shield"] = ShockAndAwe_Bars.constants["Lightning Shield Icon"]
					return true
				end
			else
				return false
			end 
		end
		
	ShockAndAwe_Bars.priorityTable.name["sl_totem"] = ShockAndAwe_Bars.constants["Stormlash Totem"]
	ShockAndAwe_Bars.priorityTable.icon["sl_totem"] = ShockAndAwe_Bars.constants["Stormlash Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["sl_totem"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Stormlash Totem"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["em"] = ShockAndAwe_Bars.constants["Elemental Mastery"]
	ShockAndAwe_Bars.priorityTable.icon["em"] = ShockAndAwe_Bars.constants["Elemental Mastery Icon"]
	ShockAndAwe_Bars.priorityTable.test["em"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Elemental Mastery"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["Ascendance"] = ShockAndAwe_Bars.constants["Ascendance"]
	ShockAndAwe_Bars.priorityTable.icon["Ascendance"] = ShockAndAwe_Bars.constants["Ascendance Icon"]
	ShockAndAwe_Bars.priorityTable.test["Ascendance"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Ascendance"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["Earth Elemental Totem"] = ShockAndAwe_Bars.constants["Earth Elemental Totem"]
	ShockAndAwe_Bars.priorityTable.icon["Earth Elemental Totem"] = ShockAndAwe_Bars.constants["Earth Elemental Totem Icon"]
	ShockAndAwe_Bars.priorityTable.test["Earth Elemental Totem"] =  
		function () 
			return ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Earth Elemental Totem"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["MW>3_LB_UF"] = L["MW>3_LB_UF"]
	ShockAndAwe_Bars.priorityTable.icon["MW>3_LB_UF"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["MW>3_LB_UF"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks >= 3 and ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Unleashed Fury"])
		end
		
	ShockAndAwe_Bars.priorityTable.name["MW>1_LB"] = L["MW>1_LB"]
	ShockAndAwe_Bars.priorityTable.icon["MW>1_LB"] = ShockAndAwe_Bars.constants["Lightning Bolt Icon"]
	ShockAndAwe_Bars.priorityTable.test["MW>1_LB"] = 
		function () 
			return ShockAndAwe_Bars.db.char.msstacks > 1 
		end
		
	ShockAndAwe_Bars.priorityTable.name["FS_UEF"] = L["FS_UEF"]
	ShockAndAwe_Bars.priorityTable.icon["FS_UEF"] = ShockAndAwe_Bars.constants["Flame Shock Icon"]
	ShockAndAwe_Bars.priorityTable.test["FS_UEF"] = 
		function () 	
		    local debuffPresent, duration = ShockAndAwe_Bars:GetDebuffInfo(ShockAndAwe_Bars.constants["Flame Shock"])
		    local ticksLeft = 6 * duration / (ShockAndAwe_Bars.db.char.priority.FSDotMax or 18)
			return (not debuffPresent or ticksLeft <= ShockAndAwe_Bars.db.char.priority.fsticksleft) and ShockAndAwe_Bars:SpellAvailable(ShockAndAwe_Bars.constants["Flame Shock"]) and ShockAndAwe_Bars:GetBuffInfo(ShockAndAwe_Bars.constants["Unleash Flame"])
		end

	ShockAndAwe_Bars.priorityTable.name["none"] = L["None"]
	_, _, ShockAndAwe_Bars.priorityTable.icon["none"] = "Interface/Tooltips/UI-Tooltip-Background"
	ShockAndAwe_Bars.priorityTable.test["none"] = function () return false end 
end
-------------------------------
-- Priority Frame functions
-------------------------------

function ShockAndAwe_Bars:SetPriorityBackdrop(icon)
--	self.PriorityFrame.Icon:SetTexture(icon)
	self.PriorityFrame:SetBackdrop({ bgFile = icon,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 12,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})
end

function ShockAndAwe_Bars:CreatePriorityFrame()
	self.updatePriority = false
	self.PriorityFrame:SetScale(ShockAndAwe_Bars.db.char.priority.scale)
	self.PriorityFrame:SetFrameStrata("BACKGROUND")
	self.PriorityFrame:SetWidth(ShockAndAwe_Bars.db.char.priority.fWidth)
	self.PriorityFrame:SetHeight(ShockAndAwe_Bars.db.char.priority.fHeight)
	self.PriorityFrame.Icon = _G["SAA_PriorityFrameIcon"]
	self:SetPriorityBackdrop("Interface/Tooltips/UI-Tooltip-Background")
	self.PriorityFrame:SetBackdropColor(0, 0, 0, ShockAndAwe_Bars.db.char.priority.alpha);
	self.PriorityFrame:SetMovable(true);
	self.PriorityFrame:RegisterForDrag("LeftButton");
	self.PriorityFrame:EnableMouse(false)
	self.PriorityFrame:SetPoint(ShockAndAwe_Bars.db.char.priority.point, ShockAndAwe_Bars.db.char.priority.relativeTo, ShockAndAwe_Bars.db.char.priority.relativePoint, ShockAndAwe_Bars.db.char.priority.xOffset, ShockAndAwe_Bars.db.char.priority.yOffset)
	self.PriorityFrame:SetScript("OnDragStart", 
		function()
			self.PriorityFrame:StartMoving();
		end );
	self.PriorityFrame:SetScript("OnDragStop",
		function()
			self.PriorityFrame:StopMovingOrSizing();
			self.PriorityFrame:SetScript("OnUpdate", nil);
			self:FinishedMoving(ShockAndAwe_Bars.db.char.priority, self.PriorityFrame);
		end );
	self:SetPriorityUpdateScript()
	if not self.PriorityFrame.topText then
		self.PriorityFrame.topText = self.PriorityFrame:CreateFontString(nil, "OVERLAY")
	end
	self.PriorityFrame.topText:SetTextColor(1,1,1,1)
	self.PriorityFrame.topText:SetFont(media:Fetch("font", ShockAndAwe_Bars.db.char.barfont), ShockAndAwe_Bars.db.char.barfontsize)
	self.PriorityFrame.topText:SetPoint("TOP", self.PriorityFrame, "TOP", 0, ShockAndAwe_Bars.db.char.barfontsize + 2)
	self.PriorityFrame.topText:SetText(string.format(L["Next Priority (Set %s)"], ShockAndAwe_Bars.db.char.priority.groupnumber))
	self.PriorityFrame.cooldown:SetAllPoints(self.PriorityFrame)
	self:CreateComboPointFrames()
	self:CreateInterruptPurgeFrames()
	
	if ShockAndAwe_Bars.db.char.priority.show and InCombatLockdown() then
		self.PriorityFrame:Show()
	else
		self.PriorityFrame:Hide()
	end
	if ShockAndAwe_Bars.db.char.priority.titleshow then
		self.PriorityFrame.topText:Show()
	else
		self.PriorityFrame.topText:Hide()
	end
	if LBF and LBFgroup then
		LBFgroup:AddButton(self.PriorityFrame)
	end
end

function ShockAndAwe_Bars:ResetPriority()
	self.PriorityFrame:ClearAllPoints()
	ShockAndAwe_Bars.db.char.priority.point = self.defaults.char.priority.point
	ShockAndAwe_Bars.db.char.priority.relativeTo = self.defaults.char.priority.relativeTo 
	ShockAndAwe_Bars.db.char.priority.relativePoint = self.defaults.char.priority.relativePoint
	ShockAndAwe_Bars.db.char.priority.xOffset = self.defaults.char.priority.xOffset
	ShockAndAwe_Bars.db.char.priority.yOffset = self.defaults.char.priority.yOffset
	ShockAndAwe_Bars.db.char.priority.fWidth = self.defaults.char.priority.fWidth
	ShockAndAwe_Bars.db.char.priority.fHeight = self.defaults.char.priority.fHeight
	ShockAndAwe_Bars.db.char.priority.scale = self.defaults.char.priority.scale
	ShockAndAwe_Bars.db.char.priority.prOption = self.defaults.char.priority.prOption
	self.PriorityFrame:SetPoint(ShockAndAwe_Bars.db.char.priority.point, ShockAndAwe_Bars.db.char.priority.relativeTo, ShockAndAwe_Bars.db.char.priority.relativePoint, ShockAndAwe_Bars.db.char.priority.xOffset, ShockAndAwe_Bars.db.char.priority.yOffset)
	self:CreatePriorityFrame()
	self:Print(L["priority_reset"])
end

function ShockAndAwe_Bars:SetPriorityUpdateScript()
	self.PriorityFrame:SetScript("OnUpdate", 
		function()
			if self.updatePriority then
				ShockAndAwe_Bars.db.char.priority.previous = ShockAndAwe_Bars.db.char.priority.next
				ShockAndAwe_Bars:SetNextPriority()
--				if ShockAndAwe_Bars.db.char.priority.next ~= ShockAndAwe_Bars.db.char.priority.previous or ShockAndAwe_Bars:CheckShamanisticRage() then
				ShockAndAwe_Bars:SetPriorityIcon(ShockAndAwe_Bars.db.char.priority.next) -- always set icon to fix the times where it was getting stuck
--				end
				if ShockAndAwe_Bars.db.char.priority.showcooldown then
					local startTime, duration = GetSpellCooldown(ShockAndAwe_Bars.constants["Lightning Shield"])
					if startTime then
						ShockAndAwe_Bars.PriorityFrame.cooldown:SetCooldown(startTime, duration)
					end
				end
			end
		end );
end

function ShockAndAwe_Bars:SetPriorityIcon(priority)
	--self:DebugPrint("setting icon to "..priority)
	if InCombatLockdown() or priority == "none" then
		local icon = self.priorityTable.icon[priority]
		if self:CheckWindShear() then
			icon = ShockAndAwe_Bars.constants["Wind Shear Icon"]
			ShockAndAwe_Bars.db.char.priority.next = "wind"
		end
		if self:CheckShamanisticRage() then
			icon = ShockAndAwe_Bars.constants["Shamanistic Rage Icon"]
			ShockAndAwe_Bars.db.char.priority.next = "mana"
		end
		-- check if we should show shamanistic rage icon
		self:SetPriorityBackdrop(icon)
	end
end

function ShockAndAwe_Bars:CheckWindShear()
	if not ShockAndAwe_Bars.db.char.priority.worldbossonly or (ShockAndAwe_Bars.db.char.priority.worldbossonly and UnitClassification("target") == "worldboss") then
		if ShockAndAwe_Bars.db.char.windshearshow and (GetNumGroupMembers()>=4) then
			-- don't use windshear option if soloing
			local _, state, _, raw = UnitDetailedThreatSituation("player", "target")
			-- state 2 or 3 = tanking which is not what an Enh Shammy should be doing
			-- state 0 threat < 100%, state 1 threat >=100%
			if not state then state = 0 end
			if not raw then raw = 0 end
			if  state > 1 or raw > ShockAndAwe_Bars.db.char.threatThreshold then
				-- we have exceeded threshold or have already started tanking so wind shear needed.
				local startTime, duration = GetSpellCooldown(ShockAndAwe_Bars.constants["Wind Shear"])
				if startTime then -- extra check in case player hasn't learned Wind Shear
					local timeleft = startTime + duration - GetTime()
					if timeleft <= ShockAndAwe_Bars.db.char.priority.cooldown then -- check its off cooldown
						return true
					end
				end
			end
		end
	end
	return false
end

function ShockAndAwe_Bars:CheckShamanisticRage()
	local srtalent
	srtalent = 1
	--_, _, _, _, srtalent = GetTalentInfo(2,26)
	if srtalent == 1 then
		local startTime, duration = GetSpellCooldown(ShockAndAwe_Bars.constants["Shamanistic Rage"])
		if startTime then 
			local timeleft = startTime + duration - GetTime()
			local manapercent = 100 * UnitMana("player") / UnitManaMax("player")
			if manapercent < ShockAndAwe_Bars.db.char.priority.srmana and timeleft <= ShockAndAwe_Bars.db.char.priority.cooldown then
				return true
			end
		end
	end
	return false
end

function ShockAndAwe_Bars:QuakingEarthEquipped()
	local rangedLink = GetInventoryItemLink("player",GetInventorySlotInfo("RangedSlot")) or "0:0"
    local _, itemID = strsplit(":", rangedLink)
	return itemID == "47667" -- returns if ranged slot has Totem of Quaking Earth
end

function ShockAndAwe_Bars:SetNextPriority()
	-- force update of stacks count to ensure priority shown correctly
	ShockAndAwe_Bars.db.char.msstacks = ShockAndAwe_Bars:GetMaelstromInfo()
	if ShockAndAwe_Bars.db.char.priority.combopoints then
		self:ShowComboPoints()
	end
	ShockAndAwe_Bars.db.char.priority.next = "none"
	for index = 1, 16 do
		if ShockAndAwe_Bars.db.char.priority.prOption[index] and self.priorityTable.test[ShockAndAwe_Bars.db.char.priority.prOption[index]] then  -- verify that the option actually exists
			if self.priorityTable.test[ShockAndAwe_Bars.db.char.priority.prOption[index]]() then
				ShockAndAwe_Bars.db.char.priority.next = ShockAndAwe_Bars.db.char.priority.prOption[index]
				return -- we need to break out of the routine as we have found the top priority
			end
		end
	end
end

function ShockAndAwe_Bars:ShowComboPoints()
	local col = ShockAndAwe_Bars.db.char.colours.maelstrom
	for index = 1, 5 do
		if self.PriorityFrame.combo[index].frame then
			if ShockAndAwe_Bars.db.char.msstacks >= index then
				self.PriorityFrame.combo[index].frame:SetBackdropColor(1, 0, 0, 1)
			else
				self.PriorityFrame.combo[index].frame:SetBackdropColor(col.r, col.g, col.b, 0)
			end
		else
			self:Print("Error could not find combo frame "..index)
		end
	end
end

function ShockAndAwe_Bars:CreateComboPointFrames()
	self.PriorityFrame.combo = {}
	for index = 1, 5 do 
		self:CreateComboPoint(index)
	end
end
	
function ShockAndAwe_Bars:CreateComboPoint(index)
	if not self.PriorityFrame.combo[index] or not self.PriorityFrame.combo[index].frame then
		self.PriorityFrame.combo[index] = {}
		self.PriorityFrame.combo[index].frame = CreateFrame("Frame", "SAA_PriorityComboFrame"..index, self.PriorityFrame)
	end
	local comboFrame = self.PriorityFrame.combo[index].frame
	local width = ShockAndAwe_Bars.db.char.priority.fWidth / 5
	local height = ShockAndAwe_Bars.db.char.priority.fHeight / 5
	comboFrame:SetScale(ShockAndAwe_Bars.db.char.priority.scale)
	comboFrame:SetFrameStrata("BACKGROUND")
	comboFrame:SetWidth(width)
	comboFrame:SetHeight(height)
	comboFrame:SetBackdrop({ bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = 4,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})	
	comboFrame:SetBackdropColor(0, 0, 0, 0);
	comboFrame:SetPoint("BOTTOMLEFT",  self.PriorityFrame, "BOTTOMLEFT", (index - 1) * width , - height)

	if ShockAndAwe_Bars.db.char.priority.combopoints then
		comboFrame:Show()
	else
		comboFrame:Hide()
	end
end

function ShockAndAwe_Bars:CreateInterruptPurgeFrames()
	if not self.PriorityFrame.interrupt then
		self.PriorityFrame.interrupt = {}
		self.PriorityFrame.interrupt.frame = CreateFrame("Frame", "SAA_PriorityInterruptFrame", self.PriorityFrame)
	end
	local interruptFrame = self.PriorityFrame.interrupt.frame
	local width = ShockAndAwe_Bars.db.char.priority.fWidth / 2
	local height = ShockAndAwe_Bars.db.char.priority.fHeight / 2
	interruptFrame:SetScale(ShockAndAwe_Bars.db.char.priority.scale)
	interruptFrame:SetFrameStrata("BACKGROUND")
	interruptFrame:SetWidth(width)
	interruptFrame:SetHeight(height)
	self:SetSubFrameBackdrop(interruptFrame, "Interface/Tooltips/UI-Tooltip-Background", 4)	
	interruptFrame:SetBackdropColor(0, 0, 0, 0);
	interruptFrame:SetPoint("BOTTOMLEFT",  self.PriorityFrame, "BOTTOMRIGHT", 0 , 0)
	if ShockAndAwe_Bars.db.char.priority.showinterrupt then
		interruptFrame:Show()
	else
		interruptFrame:Hide()
	end

	if not self.PriorityFrame.purge then
		self.PriorityFrame.purge = {}
		self.PriorityFrame.purge.frame = CreateFrame("Frame", "SAA_PrioritypurgeFrame", self.PriorityFrame)
	end
	local purgeFrame = self.PriorityFrame.purge.frame
	purgeFrame:SetScale(ShockAndAwe_Bars.db.char.priority.scale)
	purgeFrame:SetFrameStrata("BACKGROUND")
	purgeFrame:SetWidth(width)
	purgeFrame:SetHeight(height)
	self:SetSubFrameBackdrop(purgeFrame, "Interface/Tooltips/UI-Tooltip-Background", 4)	
	purgeFrame:SetBackdropColor(0, 0, 0, 0);
	purgeFrame:SetPoint("BOTTOMLEFT",  self.PriorityFrame, "BOTTOMRIGHT", 0 , height)	
	if ShockAndAwe_Bars.db.char.priority.showpurge then
		purgeFrame:Show()
	else
		purgeFrame:Hide()
	end
end

function ShockAndAwe_Bars:SetSubFrameBackdrop(subFrame, icon, edgeSize)
	subFrame:SetBackdrop({ bgFile = icon,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = false, tileSize = 0, edgeSize = edgeSize,
		insets = { left = 2, right = 2, top = 2, bottom = 2 }
	})	
end