if not ShockAndAwe_Bars then return end

local L = LibStub("AceLocale-3.0"):GetLocale("ShockAndAwe_Bars")
local media = LibStub:GetLibrary("LibSharedMedia-3.0");
local C = ShockAndAwe_Bars.constants -- Defined in Constants.LUA no locale needed.

ShockAndAwe_Bars.UptimeFrame = CreateFrame("Frame", "SAA_UptimeFrame", UIParent)

---------------------------
-- Uptime functions
---------------------------

function ShockAndAwe_Bars:InitialiseUptime()
	ShockAndAwe_Bars.uptime = {}
	ShockAndAwe_Bars.uptime = {
		session = {},
		lastfight = {},
		incombat = false,
		TimerEvent = nil,
		currentTime = nil,
	}
	self:InitialiseUptimeBuffs(ShockAndAwe_Bars.uptime.session)
	self:InitialiseUptimeBuffs(ShockAndAwe_Bars.uptime.lastfight)
end

function ShockAndAwe_Bars:InitialiseUptimeBuffs(timeframe)
	timeframe.totalTime = 0
	if not timeframe.buffs then
		timeframe.buffs = {}
	end
	if not timeframe.buffs[ShockAndAwe_Bars.constants["Flurry"]] then
		timeframe.buffs[ShockAndAwe_Bars.constants["Flurry"]] = {}
		timeframe.buffs[ShockAndAwe_Bars.constants["Flurry"]].name = ShockAndAwe_Bars.constants["Flurry"]
		timeframe.buffs[ShockAndAwe_Bars.constants["Flurry"]].icon = ShockAndAwe_Bars.constants["Flurry Icon"]
	end
	timeframe.buffs[ShockAndAwe_Bars.constants["Flurry"]].uptime = 0
end

function ShockAndAwe_Bars:UpdateUptime()
	if not ShockAndAwe_Bars.uptime.incombat then
		return
	end
	local currentTime = GetTime()
	-- add conditional safety measure - we should never have no current time but to be sure...
	if  not ShockAndAwe_Bars.uptime.currentTime then
		ShockAndAwe_Bars.uptime.currentTime = currentTime
	end
	local diffTime = currentTime - ShockAndAwe_Bars.uptime.currentTime
	
	ShockAndAwe_Bars.uptime.session.totalTime = ShockAndAwe_Bars.uptime.session.totalTime + diffTime;
	ShockAndAwe_Bars.uptime.lastfight.totalTime = ShockAndAwe_Bars.uptime.lastfight.totalTime + diffTime;
	
	local i = 1
	local buff = UnitBuff("player", i)
	while buff do
		if ShockAndAwe_Bars.uptime.session.buffs[buff] then
			ShockAndAwe_Bars.uptime.session.buffs[buff].uptime = ShockAndAwe_Bars.uptime.session.buffs[buff].uptime + diffTime
		end
		if ShockAndAwe_Bars.uptime.lastfight.buffs[buff] then
			ShockAndAwe_Bars.uptime.lastfight.buffs[buff].uptime = ShockAndAwe_Bars.uptime.lastfight.buffs[buff].uptime + diffTime
		end		
		i = i + 1
		buff = UnitBuff("player", i)
	end	
	ShockAndAwe_Bars.uptime.currentTime = currentTime
	self:UpdateUptimeFrames(false)
end

function ShockAndAwe_Bars:CreateUptimeFrame()
	ShockAndAwe_Bars.UptimeFrame:ClearAllPoints()
	ShockAndAwe_Bars.UptimeFrame:SetScale(ShockAndAwe_Bars.db.char.uptime.scale)
	ShockAndAwe_Bars.UptimeFrame:SetFrameStrata("BACKGROUND")
	ShockAndAwe_Bars.UptimeFrame:SetWidth(ShockAndAwe_Bars.db.char.uptime.fWidth + ShockAndAwe_Bars.db.char.uptime.barHeight)
	ShockAndAwe_Bars.UptimeFrame:SetHeight(ShockAndAwe_Bars.db.char.uptime.fHeight + 16)
	ShockAndAwe_Bars.UptimeFrame:SetBackdrop(self.frameBackdrop)
	ShockAndAwe_Bars.UptimeFrame:SetBackdropColor(0, 0, 0, ShockAndAwe_Bars.db.char.uptime.alpha)
	ShockAndAwe_Bars.UptimeFrame:SetMovable(true)
	ShockAndAwe_Bars.UptimeFrame:RegisterForDrag("LeftButton")
	ShockAndAwe_Bars.UptimeFrame:SetPoint(ShockAndAwe_Bars.db.char.uptime.point, ShockAndAwe_Bars.db.char.uptime.relativeTo, ShockAndAwe_Bars.db.char.uptime.relativePoint, ShockAndAwe_Bars.db.char.uptime.xOffset, ShockAndAwe_Bars.db.char.uptime.yOffset)
	ShockAndAwe_Bars.UptimeFrame:SetScript("OnDragStart", 
		function()
			ShockAndAwe_Bars.UptimeFrame:StartMoving();
		end );
	ShockAndAwe_Bars.UptimeFrame:SetScript("OnDragStop",
		function()
			ShockAndAwe_Bars.UptimeFrame:StopMovingOrSizing();
			ShockAndAwe_Bars.UptimeFrame:SetScript("OnUpdate", nil);
			self:FinishedMoving(ShockAndAwe_Bars.db.char.uptime, ShockAndAwe_Bars.UptimeFrame);
		end );
	
	local barfont = media:Fetch("font", ShockAndAwe_Bars.db.char.barfont)
	if not ShockAndAwe_Bars.UptimeFrame.topText then
		ShockAndAwe_Bars.UptimeFrame.topText = ShockAndAwe_Bars.UptimeFrame:CreateFontString(nil, "OVERLAY")
	end
	ShockAndAwe_Bars.UptimeFrame.topText:SetTextColor(1,1,1,1)
	ShockAndAwe_Bars.UptimeFrame.topText:SetFont(barfont, ShockAndAwe_Bars.db.char.barfontsize, ShockAndAwe_Bars.db.char.barfonteffect)
	ShockAndAwe_Bars.UptimeFrame.topText:SetPoint("TOP", ShockAndAwe_Bars.UptimeFrame, "TOP", 0, - 2)
	ShockAndAwe_Bars.UptimeFrame.topText:SetText(L["uptime_session"])
	
	if not ShockAndAwe_Bars.UptimeFrame.ResetButton then
		ShockAndAwe_Bars.UptimeFrame.ResetButton=CreateFrame("Button", nil, ShockAndAwe_Bars.UptimeFrame)
	end
	local size = ShockAndAwe_Bars.UptimeFrame.topText:GetHeight()
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetNormalTexture("Interface\\Addons\\ShockAndAwe_Bars\\textures\\icon-reset")
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp")
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetWidth(size)
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetHeight(size)
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetPoint("TOPRIGHT", ShockAndAwe_Bars.UptimeFrame.topText, "TOPLEFT", -5, 0)
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetScript("OnClick",function() ShockAndAwe_Bars:ResetUptimeValues() end)
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetFrameLevel(ShockAndAwe_Bars.UptimeFrame.ResetButton:GetFrameLevel()+1)
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetScript("OnEnter", function() UptimeResetButtonTooltip() end);
	ShockAndAwe_Bars.UptimeFrame.ResetButton:SetScript("OnLeave", function() GameTooltip:Hide() end);


	local baseOffset = (-1 * ShockAndAwe_Bars.db.char.uptime.fHeight / 6) - 2 -- 6 bars + top line + middle line 
	if not ShockAndAwe_Bars.UptimeFrame.middleText then
		ShockAndAwe_Bars.UptimeFrame.middleText = ShockAndAwe_Bars.UptimeFrame:CreateFontString(nil, "OVERLAY")
	end
	ShockAndAwe_Bars.UptimeFrame.middleText:SetTextColor(1,1,1,1)
	ShockAndAwe_Bars.UptimeFrame.middleText:SetFont(barfont, ShockAndAwe_Bars.db.char.barfontsize, ShockAndAwe_Bars.db.char.barfonteffect)
	ShockAndAwe_Bars.UptimeFrame.middleText:SetPoint("TOP", ShockAndAwe_Bars.UptimeFrame, "TOP", 0, 3 * baseOffset - 2)
	ShockAndAwe_Bars.UptimeFrame.middleText:SetText(L["uptime_lastfight"])
	if ShockAndAwe_Bars.db.char.uptime.show then
		ShockAndAwe_Bars.UptimeFrame:Show()
	else
		ShockAndAwe_Bars.UptimeFrame:Hide()
	end
	local session = ShockAndAwe_Bars.uptime.session
	local lastfight = ShockAndAwe_Bars.uptime.lastfight
	
	self:CreateUptimeBarFrame(session.buffs[ShockAndAwe_Bars.constants["Flurry"]], ShockAndAwe_Bars.db.char.uptime.flurry, baseOffset)
	
	self:CreateUptimeBarFrame(lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]], ShockAndAwe_Bars.db.char.uptime.flurry, 4 * baseOffset)
	
end

function ShockAndAwe_Bars:CreateUptimeBarFrame(buff, buffColours, frameOffset)
	if not buff.barFrame then
		buff.barFrame = CreateFrame("Frame", nil, ShockAndAwe_Bars.UptimeFrame)
	end
	buff.barFrame:SetFrameStrata("LOW")
	buff.barFrame:SetWidth(ShockAndAwe_Bars.db.char.uptime.barWidth + ShockAndAwe_Bars.db.char.uptime.barHeight)
	buff.barFrame:SetHeight(ShockAndAwe_Bars.db.char.uptime.barHeight)
	buff.barFrame:ClearAllPoints()
	buff.barFrame:SetPoint("TOPLEFT", ShockAndAwe_Bars.UptimeFrame, "TOPLEFT", ShockAndAwe_Bars.db.char.uptime.barHeight, frameOffset)
	buff.barFrame:SetScript("OnUpdate", function() self:UpdateUptimeFrames(false); end );
	
	if not buff.barFrame.buffIcon then
		buff.barFrame.buffIcon = CreateFrame("Frame", nil, buff.barFrame)
	end
	buff.barFrame.buffIcon:SetWidth(ShockAndAwe_Bars.db.char.uptime.barHeight)
	buff.barFrame.buffIcon:SetHeight(ShockAndAwe_Bars.db.char.uptime.barHeight)
	buff.barFrame.buffIcon:SetBackdrop({ bgFile = buff.icon })
	buff.barFrame.buffIcon:SetPoint("TOPRIGHT", buff.barFrame, "TOPLEFT", 12, 0)
	
	if not buff.barFrame.statusbar then
		buff.barFrame.statusbar = CreateFrame("StatusBar", nil, buff.barFrame, "TextStatusBar")
	end
	buff.barFrame.statusbar:ClearAllPoints()
	buff.barFrame.statusbar:SetHeight(ShockAndAwe_Bars.db.char.uptime.barHeight)
	buff.barFrame.statusbar:SetWidth(ShockAndAwe_Bars.db.char.uptime.barWidth)
	buff.barFrame.statusbar:SetPoint("RIGHT", buff.barFrame, "RIGHT")
	buff.barFrame.statusbar:SetStatusBarTexture(media:Fetch("statusbar", ShockAndAwe_Bars.db.char.texture))
	buff.barFrame.statusbar:SetStatusBarColor(buffColours.r, buffColours.g, buffColours.b, buffColours.a)
	buff.barFrame.statusbar:SetMinMaxValues(0,1)
	buff.barFrame.statusbar:SetValue(0)
	
	local barfont = media:Fetch("font", ShockAndAwe_Bars.db.char.barfont)
	if not buff.barFrame.statusbar.leftText then
		buff.barFrame.statusbar.leftText = buff.barFrame.statusbar:CreateFontString(nil, "OVERLAY")
	end
	buff.barFrame.statusbar.leftText:SetTextColor(1,1,1,1)
	buff.barFrame.statusbar.leftText:SetFont(barfont, ShockAndAwe_Bars.db.char.barfontsize, ShockAndAwe_Bars.db.char.barfonteffect)
	buff.barFrame.statusbar.leftText:SetPoint("TOPLEFT", buff.barFrame.statusbar, "TOPLEFT")
	buff.barFrame.statusbar.leftText:SetText("0.0s")
	
	if not buff.barFrame.statusbar.rightText then
		buff.barFrame.statusbar.rightText = buff.barFrame.statusbar:CreateFontString(nil, "OVERLAY")
	end
	buff.barFrame.statusbar.rightText:SetTextColor(1,1,1,1)
	buff.barFrame.statusbar.rightText:SetFont(barfont, ShockAndAwe_Bars.db.char.barfontsize, ShockAndAwe_Bars.db.char.barfonteffect)
	buff.barFrame.statusbar.rightText:SetPoint("TOPRIGHT", buff.barFrame.statusbar, "TOPRIGHT")
	buff.barFrame.statusbar.rightText:SetText("0%")
	
	buff.barFrame.statusbar:SetScript("OnEnter", function() UptimeStatusBarTooltip(buff) end);
	buff.barFrame.statusbar:SetScript("OnLeave", function() GameTooltip:Hide() end);
	buff.barFrame.buffIcon:SetScript("OnEnter", function() UptimeStatusBarTooltip(buff) end);
	buff.barFrame.buffIcon:SetScript("OnLeave", function() GameTooltip:Hide() end);
end

function UptimeStatusBarTooltip(buff)
	if not InCombatLockdown() then
		GameTooltip:SetOwner(buff.barFrame, "ANCHOR_BOTTOMLEFT");
		GameTooltip:AddLine(buff.name)
		GameTooltip:AddTexture(buff.icon)
		GameTooltip:AddLine((L["Uptime is %s (%s)"]):format(
				buff.barFrame.statusbar.leftText:GetText(),
				buff.barFrame.statusbar.rightText:GetText()
			))
		GameTooltip:Show()
	else
		GameTooltip:Hide()
	end
end

function UptimeResetButtonTooltip()
	if not InCombatLockdown() then
		GameTooltip:SetOwner(ShockAndAwe_Bars.UptimeFrame.ResetButton, "ANCHOR_BOTTOMLEFT");
		GameTooltip:AddLine(L["Reset Session statistics"])
		GameTooltip:Show()
	else
		GameTooltip:Hide()
		
	end
end

function ShockAndAwe_Bars:UpdateUptimeFrames(forceUpdate)
	if InCombatLockdown() or forceUpdate then
		local session = ShockAndAwe_Bars.uptime.session
		local lastfight = ShockAndAwe_Bars.uptime.lastfight
		session.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.leftText:SetText(self:DurationString(session.buffs[ShockAndAwe_Bars.constants["Flurry"]].uptime))
		
		lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.leftText:SetText(self:DurationString(lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].uptime))
		

		local percent
		if session.totalTime == 0 then
			session.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar:SetValue(0)
			session.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.rightText:SetText("0%")
		else
			percent = session.buffs[ShockAndAwe_Bars.constants["Flurry"]].uptime / session.totalTime
			session.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar:SetValue(percent)
			session.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.rightText:SetText((("%d"):format(percent * 100)) .. "%")
		end

		if lastfight.totalTime == 0 then
			lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar:SetValue(0)
			lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.rightText:SetText("0%")
		else
			percent = lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].uptime / lastfight.totalTime
			lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar:SetValue(percent)
			lastfight.buffs[ShockAndAwe_Bars.constants["Flurry"]].barFrame.statusbar.rightText:SetText((("%d"):format(percent * 100)) .. "%")
			
		end
	end
end

function ShockAndAwe_Bars:ResetUptime()
	ShockAndAwe_Bars.UptimeFrame:ClearAllPoints()
	ShockAndAwe_Bars.db.char.uptime.point = self.defaults.char.uptime.point
	ShockAndAwe_Bars.db.char.uptime.relativeTo = self.defaults.char.uptime.relativeTo 
	ShockAndAwe_Bars.db.char.uptime.relativePoint = self.defaults.char.uptime.relativePoint
	ShockAndAwe_Bars.db.char.uptime.xOffset = self.defaults.char.uptime.xOffset
	ShockAndAwe_Bars.db.char.uptime.yOffset = self.defaults.char.uptime.yOffset
	ShockAndAwe_Bars.db.char.uptime.fWidth = self.defaults.char.uptime.fWidth
	ShockAndAwe_Bars.db.char.uptime.fHeight = self.defaults.char.uptime.fHeight
	ShockAndAwe_Bars.db.char.uptime.scale = self.defaults.char.uptime.scale
	self:CreateUptimeFrame()
	self:Print(L["uptime_reset"])
end

function ShockAndAwe_Bars:ResetUptimeValues()
	self:InitialiseUptimeBuffs(ShockAndAwe_Bars.uptime.session)
	self:InitialiseUptimeBuffs(ShockAndAwe_Bars.uptime.lastfight)
	self:UpdateUptimeFrames(true)
end